# -*- coding: utf-8 -*-
"""
The imagestack module contains the base functionality for nD bio image opening with metadata support.
"""

from .imagestack import ImageStack, Dimensions, FloatFilter, MinMaxFilter, UnwrapFilter

