# -*- coding: utf-8 -*-
"""
pilyso ... python image analysis software

This library has devised to serve as base for multiple image analysis tools.
So far it is not completely ready as a stand-alone library,
and as such just bundled with mycelyso as-of now.
"""
